"""
Created on 2021

@author: Eduardo Vitral
"""

###############################################################################
#
# June 2021, Paris
#
# This file contains the main functions concerning the dispersion functions,
# (i.e., velocity disperstion and anisotropy). It also converts plane of sky
# velocities (and uncertainties) from (RA,Dec) to polar coordinates.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

from . import position
from . import angle

import numpy as np
import operator
from scipy.interpolate import PchipInterpolator
import math
from scipy.spatial import ConvexHull
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
from functools import partial

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Proper motions and conversions"
# ------------------------------------------------------------------------------


def v_sky_to_polar(a, d, pma, pmd, a0, d0, pma0, pmd0):
    """
    Transforms proper motions in RA Dec into polar coordinates
    (radial and tangential).

    Parameters
    ----------
    a : array_like
        RA of the source.
    d : array_like
        Dec of the source.
    pma : array_like
        PMRA of the source.
    pmd : array_like
        PMDec of the source.
    a0 : float
        Bulk RA.
    d0 : float
        Bulk Dec.
    pma0 : float
        Bulk PMRA.
    pmd0 : float
        Bulk PMDec.

    Returns
    -------
    pmr : array_like
        PM in radial direction of the source.
    pmt : array_like
        PM in tangential direction of the source.

    """

    a = a * np.pi / 180
    d = d * np.pi / 180
    a0 = a0 * np.pi / 180
    d0 = d0 * np.pi / 180

    sinda = np.sin(a - a0)
    cosda = np.cos(a - a0)
    sind = np.sin(d)
    sind0 = np.sin(d0)
    cosd = np.cos(d)
    cosd0 = np.cos(d0)

    dx = sinda * cosd
    dy = cosd0 * sind - sind0 * cosd * cosda
    rho = np.sqrt(dx * dx + dy * dy)
    theta = np.arccos(sind0 * sind + cosd0 * cosd * cosda)

    cost = np.cos(theta)

    dmux = cosda * (pma - pma0 * cosd / cosd0) - sind * sinda * pmd
    dmuy = (
        (cosd * cosd0 + sind * sind0 * cosda) * pmd
        - cost * pmd0
        + (pma - pma0 * cosd / cosd0) * sind0 * sinda
    )

    pmr = (dx * dmux + dy * dmuy) / rho
    pmt = (-dx * dmuy + dy * dmux) / rho

    return pmr, pmt


def unc_sky_to_polar(a, d, epma, epmd, epmad, a0, d0, epma0, epmd0):
    """
    Transforms proper motions uncertainties in RA Dec into polar coordinates
    uncertainties (radial and tangential).

    Parameters
    ----------
    a : array_like
        RA of the source.
    d : array_like
        Dec of the source.
    epma : array_like
        Uncertainty in PMRA of the source.
    epmd : array_like
        Uncertainty in PMDec of the source.
    epmad : array_like
        Correlation between epma and epmd.
    a0 : float
        Bulk RA.
    d0 : float
        Bulk Dec.
    epma0 : float
        Uncertainty in Bulk PMRA.
    epmd0 : float
        Uncertainty in Bulk PMDec.

    Returns
    -------
    uncpmr : array_like
        Uncertainty in PM in radial direction.
    uncpmt : array_like
        Uncertainty in PM in tangential direction.

    """

    a = a * np.pi / 180
    d = d * np.pi / 180
    a0 = a0 * np.pi / 180
    d0 = d0 * np.pi / 180

    sina = np.sin(a)
    cosa = np.cos(a)
    sina0 = np.sin(a0)
    cosa0 = np.cos(a0)
    sind = np.sin(d)
    cosd = np.cos(d)
    sind0 = np.sin(d0)
    cosd0 = np.cos(d0)
    sinda = np.sin(a - a0)
    cosda = np.cos(a - a0)

    dentheta = np.sqrt(
        cosd ** 2 * sinda ** 2 + (cosd0 * sind - cosda * cosd * sind0) ** 2
    )

    dvdpma = (cosd0 * sinda * (cosda * cosd * cosd0 + sind * sind0)) / dentheta
    dvdpmd = (
        -cosd * sinda ** 2 * sind
        + (cosd0 * sind - cosda * cosd * sind0) * (cosd * cosd0 + cosda * sind * sind0)
    ) / dentheta
    dvdpma0 = -cosd * sinda * (cosda * cosd * cosd0 + sind * sind0) / dentheta
    dvdpmd0 = (
        -(cosd0 * sind - cosda * cosd * sind0)
        * (cosda * cosd * cosd0 + sind * sind0)
        / dentheta
    )

    uncpmr = np.sqrt(
        (dvdpma * epma) ** 2
        + (dvdpmd * epmd) ** 2
        + (dvdpma0 * epma0) ** 2
        + (dvdpmd0 * epmd0) ** 2
        + 2 * (dvdpma * dvdpmd * epma * epmd * epmad)
    )

    dvdpma = (
        cosa * cosa0 * cosd0 * sind + cosd0 * sina * sina0 * sind - cosd * sind0
    ) / dentheta
    dvdpmd = (-cosd0 * sinda) / dentheta
    dvdpma0 = (
        cosd
        * (-cosa * cosa0 * sind - sina * sina0 * sind + cosd * sind0 / cosd0)
        / dentheta
    )
    dvdpmd0 = cosd * sinda * (cosda * cosd * cosd0 + sind * sind0) / dentheta

    uncpmt = np.sqrt(
        (dvdpma * epma) ** 2
        + (dvdpmd * epmd) ** 2
        + (dvdpma0 * epma0) ** 2
        + (dvdpmd0 * epmd0) ** 2
        + 2 * (dvdpma * dvdpmd * epma * epmd * epmad)
    )

    return uncpmr, uncpmt


def pmr_corr(vlos, r, d):
    """
    Correction on radial proper motion due to apparent contraction/expansion
    of the cluster.

    Parameters
    ----------
    vlos : float
        Line of sight velocity, in km/s.
    r : array_like, float
        Projected radius, in degrees.
    d : float
        Cluster distance from the Sun, in kpc.

    Returns
    -------
    pmr : array_like, float
        Correction in the radial component of the proper motion, in mas/yr.

    """
    r = r * 60
    # Equation 4 from Bianchini et al. 2018.
    pmr = -6.1363 * 1e-5 * vlos * r / d

    return pmr


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Dispersion functions"
# ------------------------------------------------------------------------------


def aux_disp(idx, y, ey, dimy, robust_sig):
    """
    Auxiliary function used by hexbin to compute the dispersion.

    Parameters
    ----------
    idx : array_like
        Array of indexes to consider in the data.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    robust_sig : boolean
        True if the user wants to compute  a dispersion
        less sensible to outliers.

    Returns
    -------
    disp : array_like
        Dispersion over the selected values.

    """

    disp = np.zeros((dimy, 1))

    if robust_sig is True:

        for i in range(0, dimy):

            disp[i, 0] = np.median(np.abs(y[i][idx] - np.median(y[i][idx]))) / 0.6745
            disp[i, 0] = np.sqrt(disp[i, 0] ** 2 - np.nanmean(ey[i][idx] ** 2))

    else:

        for i in range(0, dimy):

            if len(idx) < 30:
                disp[i, 0] = np.sqrt(
                    np.nanstd(y[i][idx], ddof=1) ** 2 - np.nanmean(ey[i][idx] ** 2)
                )
            else:
                disp[i, 0] = np.sqrt(
                    np.nanstd(y[i][idx], ddof=0) ** 2 - np.nanmean(ey[i][idx] ** 2)
                )

    disp = np.sqrt(np.sum(disp * disp, axis=0)) / np.sqrt(dimy)

    return disp


def aux_err(idx, y, ey, dimy, robust_sig, bootp):
    """
    Auxiliary function used by hexbin to compute the dispersion.

    Parameters
    ----------
    idx : array_like
        Array of indexes to consider in the data.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    robust_sig : boolean
        True if the user wants to compute  a dispersion
        less sensible to outliers.
    bootp : boolean
        True if the errors are drawn from a Bootstrap method.

    Returns
    -------
    err : array_like
        Uncertainty on the dispersion over the selected values.

    """

    disp = np.zeros((dimy, 1))
    err = np.zeros((dimy, 1))

    if robust_sig is True:

        for i in range(0, dimy):

            disp[i, 0] = np.median(np.abs(y[i][idx] - np.median(y[i][idx]))) / 0.6745
            disp[i, 0] = np.sqrt(disp[i, 0] ** 2 - np.nanmean(ey[i][idx] ** 2))

            if bootp is True:
                err[i, 0] = bootstrap(y[i][idx])
            else:
                err[i, 0] = disp[i, 0] / np.sqrt(2 * (len(y[i][idx]) - 1))

    else:

        for i in range(0, dimy):

            if len(idx) < 30:
                disp[i, 0] = np.sqrt(
                    np.nanstd(y[i][idx], ddof=1) ** 2 - np.nanmean(ey[i][idx] ** 2)
                )
            else:
                disp[i, 0] = np.sqrt(
                    np.nanstd(y[i][idx], ddof=0) ** 2 - np.nanmean(ey[i][idx] ** 2)
                )

            if bootp is True:
                err[i, 0] = bootstrap(y[i][idx])
            else:
                err[i, 0] = disp[i, 0] / np.sqrt(2 * (len(y[i][idx]) - 1))

    err = np.sqrt(np.nansum(err * err, axis=0)) / np.sqrt(dimy)

    return err


def bin_disp1d(x, y, ey, dimy, bins, bootp=True, logx=True):
    """
    Computes the dispersion of y (and its uncertainty) as a function of x.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bins : int
        Number of bins.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is True.

    Returns
    -------
    r : array_like
        Binned version of x.
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    disp = np.zeros((dimy, bins - 1))
    err = np.zeros((dimy, bins - 1))
    r = np.zeros(bins - 1)

    if logx is True:
        rbin = np.logspace(np.log10(np.nanmin(x)), np.log10(np.nanmax(x)), bins)
    else:
        rbin = np.linspace(np.nanmin(x), np.nanmax(x), bins)

    for i in range(0, dimy):
        for j in range(0, bins - 1):
            cond1 = np.where(x < rbin[j + 1])
            cond2 = np.where(x >= rbin[j])
            cond = np.intersect1d(cond1, cond2)
            if len(cond) > 30:
                dof = 0
            else:
                dof = 1

            if logx is True:
                r[j] = np.sqrt(rbin[j] * rbin[j + 1])
            else:
                r[j] = 0.5 * (rbin[j] * rbin[j + 1])

            disp[i, j] = np.sqrt(
                np.nanstd(y[i][cond], ddof=dof) ** 2 - np.nanmean(ey[i][cond] ** 2)
            )

            if bootp is True:
                err[i, j] = bootstrap(y[i][cond])
            else:
                err[i, j] = disp[i, j] / np.sqrt(2 * (len(y[i][cond]) - 1))

    disp = np.sqrt(np.sum(disp * disp, axis=0)) / np.sqrt(dimy)
    err = np.sqrt(np.sum(err * err, axis=0)) / np.sqrt(dimy)

    return r, disp, err


def moving_grid1d(x, y, ey, dimy, bootp=True, logx=True, bins=10, ngrid=10):
    """
    Calculates the dispersion in a moving grid.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    bins : int, optional
        Number of bins.
        The default is None.
    ngrid : int, optional
        Number of grids per bin.
        The default is 10.

    Returns
    -------
    r : array_like
        Binned version of x.
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    if bins is None:
        bins = int(0.5 * position.good_bin(x))

    if logx is True:
        rbin = np.logspace(np.log10(np.nanmin(x)), np.log10(np.nanmax(x)), bins + 1)
        rbin = np.log10(rbin)
    else:
        rbin = np.linspace(np.nanmin(x), np.nanmax(x), bins + 1)

    for i in range(1, ngrid):

        add_x = (rbin[1] - rbin[0]) * i / (1 + ngrid)

        if logx is True:
            x_ini = 10 ** (rbin[0] + add_x)
            x_fin = 10 ** (rbin[len(rbin) - 2] + add_x)
        else:
            x_ini = rbin[0] + add_x
            x_fin = rbin[len(rbin) - 2] + add_x

        idx_x = np.intersect1d(np.where(x >= x_ini), np.where(x < x_fin))
        idx_x = idx_x.astype(int)
        xi = x[idx_x]
        yi = y[:, idx_x]
        eyi = ey[:, idx_x]

        if i == 1:
            r, disp, err = bin_disp1d(
                xi, yi, eyi, dimy, bins=bins, bootp=bootp, logx=logx
            )
        else:
            ri, dispi, erri = bin_disp1d(
                xi, yi, eyi, dimy, bins=bins, bootp=bootp, logx=logx
            )

            r = np.append(r, ri)
            disp = np.append(disp, dispi)
            err = np.append(err, erri)

    r = np.asarray(r)
    disp = np.asarray(disp)
    err = np.asarray(err)

    # Sorts the values according to R_proj
    L = sorted(zip(r, disp, err), key=operator.itemgetter(0))
    r, disp, err = zip(*L)

    r = np.asarray(r)
    disp = np.asarray(disp)
    err = np.asarray(err)

    return r, disp, err


def equal_size(x, y, ey, dimy, bootp=True, logx=True, nbin=10):
    """
    Calculates the dispersion in a fixed bin size grid.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    nbin : int, optional
        Number of tracers per bin. The default is 10.

    Returns
    -------
    r : array_like
        Binned version of x.
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    if nbin > 30:
        dof = 0
    else:
        dof = 1

    size = math.floor(len(x) / nbin)
    disp = np.zeros((dimy, size))
    err = np.zeros((dimy, size))
    r = np.zeros(size)

    for i in range(0, size):
        if logx is True:
            r[i] = np.sqrt(x[nbin * (1 + i) - 1] * x[nbin * i])
        else:
            r[i] = (x[nbin * (1 + i) - 1] + x[nbin * i]) * 0.5

        for j in range(0, dimy):

            disp[j, i] = np.sqrt(
                np.nanstd(y[j][nbin * i : nbin * (i + 1) - 1], ddof=dof) ** 2
                - np.nanmean(ey[j][nbin * i : nbin * (i + 1) - 1] ** 2)
            )

            if bootp is True:
                err[j, i] = bootstrap(y[j][nbin * i : nbin * (i + 1) - 1])
            else:
                err[j, i] = disp[j, i] / np.sqrt(
                    2
                    * (
                        len(
                            y[j][nbin * i : nbin * (i + 1) - 1][
                                np.logical_not(
                                    np.isnan(y[j][nbin * i : nbin * (i + 1) - 1])
                                )
                            ]
                        )
                        - 1
                    )
                )

    disp = np.sqrt(np.sum(disp * disp, axis=0)) / np.sqrt(dimy)
    err = np.sqrt(np.sum(err * err, axis=0)) / np.sqrt(dimy)

    return r, disp, err


def perc_bins(x, y, ey, dimy, bootp=True, logx=True, nnodes=5):
    """
    Calculates the dispersion in a grid whose size varies geometrically with
    the parcentile of x.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    nnodes : int, optional
        Number of grids. The default is 5.

    Returns
    -------
    r : array_like
        Binned version of x.
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    size = nnodes
    disp = np.zeros((dimy, size))
    err = np.zeros((dimy, size))
    r = np.zeros(size)

    for i in range(0, nnodes):

        z = 100 ** (1 / nnodes)
        qi = max(0, 0.01 * z ** i)
        qf = min(0.01 * z ** (i + 1), 1)

        ri = position.quantile(x, qi)
        rf = position.quantile(x, qf)

        idxr = np.intersect1d(np.where(x >= ri), np.where(x <= rf))

        if len(idxr) > 30:
            dof = 0
        else:
            dof = 1

        if logx is True:
            r[i] = np.sqrt(np.nanmin(x[idxr]) * np.nanmax(x[idxr]))
        else:
            r[i] = (np.nanmin(x[idxr]) + np.nanmax(x[idxr])) * 0.5

        for j in range(0, dimy):

            disp[j, i] = np.sqrt(
                np.nanstd(y[j][idxr], ddof=dof) ** 2 - np.nanmean(ey[j][idxr] ** 2)
            )

            if bootp is True:
                err[j, i] = bootstrap(y[j][idxr])
            else:
                err[j, i] = disp[j, i] / np.sqrt(
                    2 * (len(y[j][idxr][np.logical_not(np.isnan(y[j][idxr]))]) - 1)
                )

    disp = np.sqrt(np.sum(disp * disp, axis=0)) / np.sqrt(dimy)
    err = np.sqrt(np.sum(err * err, axis=0)) / np.sqrt(dimy)

    return r, disp, err


def closest_points(x, y, ey, dimy, bootp=True, logx=True, nbin=5):
    """
    Calculates the dispersion in a grid of the closest nbin tracers.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    nbin : int, optional
        Number of closest points. The default is 5.

    Returns
    -------
    r : array_like
        Binned version of x.
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    size = len(x)
    disp = np.zeros((dimy, size))
    err = np.zeros((dimy, size))
    r = np.zeros(size)

    for i in range(0, size):

        dist_r = np.abs(x[i] - x)
        idxr = (np.argpartition(dist_r, nbin)[:nbin]).astype(int)

        if len(idxr) > 30:
            dof = 0
        else:
            dof = 1

        r[i] = x[i]

        for j in range(0, dimy):

            disp[j, i] = np.sqrt(
                np.nanstd(y[j][idxr], ddof=dof) ** 2 - np.nanmean(ey[j][idxr] ** 2)
            )

            if bootp is True:
                err[j, i] = bootstrap(y[j][idxr])
            else:
                err[j, i] = disp[j, i] / np.sqrt(
                    2 * (len(y[j][idxr][np.logical_not(np.isnan(y[j][idxr]))]) - 1)
                )

    disp = np.sqrt(np.sum(disp * disp, axis=0)) / np.sqrt(dimy)
    err = np.sqrt(np.sum(err * err, axis=0)) / np.sqrt(dimy)

    return r, disp, err


def disp1d(
    x,
    y,
    ey,
    dimy,
    bins="percentile",
    smooth=True,
    bootp=True,
    logx=True,
    nbin=None,
    polorder=None,
    return_fits=False,
):
    """
    Computes the dispersion of y.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    bins : int, string, optional
        Number of bins or method used to bin the data.
        "moving" stands for a moving grid, later interpolated with a
        cubic spline.
        The default is "percentile".
    smooth : boolean, optional
        True if the dispersion calculated from the "moving" method should
        be smoothed.
        The default is True.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    nbin : Auxiliar value for binning when bins is not an integer.
        The default is None.
    polorder : int, optional
        Order of smoothing polynomial.
        The default is None.
    return_fits: Whether the user wants to return the polynomial
        smoothing fits.
        The default is False.

    Returns
    -------
    r : array_like
        Reference array (possibly binned).
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    if isinstance(bins, int) is True:
        r, disp, err = bin_disp1d(x, y, ey, dimy, bins=bins, bootp=bootp, logx=logx)

    if bins == "moving":
        if nbin is None:
            bins = int(0.5 * position.good_bin(x))
        else:
            bins = nbin
        ngrid = 2
        r, disp, err = moving_grid1d(
            x, y, ey, dimy, bootp=bootp, logx=logx, bins=bins, ngrid=ngrid
        )

    if bins == "fix-size":
        if nbin is None:
            bins = int(position.good_bin(x))
            nbin = int(len(x) / bins)
        else:
            nbin = nbin
        r, disp, err = equal_size(x, y, ey, dimy, bootp=bootp, logx=logx, nbin=nbin)

    if bins == "percentile":
        if nbin is None:
            nnodes = int(2 * position.good_bin(x))
        else:
            nnodes = nbin
        r, disp, err = perc_bins(x, y, ey, dimy, bootp=bootp, logx=logx, nnodes=nnodes)

    if bins == "closest":
        if nbin is None:
            bins = int(position.good_bin(x))
            nbin = int(len(x) / bins)
        else:
            nbin = nbin
        r, disp, err = closest_points(x, y, ey, dimy, bootp=bootp, logx=logx, nbin=nbin)

    nonan1 = np.logical_not(np.isnan(disp))
    nonan2 = np.logical_not(np.isnan(err))
    nonan = nonan1 * nonan2

    rmin = np.nanmin(r)
    rmax = np.nanmax(r)
    idxrange = np.intersect1d(np.where(x > rmin), np.where(x < rmax))

    if smooth is True:

        disp = PchipInterpolator(r[nonan], disp[nonan])(x[idxrange])
        err = PchipInterpolator(r[nonan], err[nonan])(x[idxrange])
        r = x[idxrange]

        if polorder is None:
            pold = int(0.2 * position.good_bin(disp))
            pole = int(0.2 * position.good_bin(err))
        else:
            pold = polorder
            pole = polorder

        if logx is False:
            poly_disp = np.polyfit(r, disp, pold)
            poly_err = np.polyfit(r, err, pole)

            disp = np.poly1d(poly_disp)(x[np.where(x < rmax)])
            err = np.poly1d(poly_err)(x[np.where(x < rmax)])
            r = x[np.where(x < rmax)]
        else:
            poly_disp = np.polyfit(np.log10(r), disp, pold)
            poly_err = np.polyfit(np.log10(r), err, pole)

            disp = np.poly1d(poly_disp)(np.log10(x[idxrange]))
            err = np.poly1d(poly_err)(np.log10(x[idxrange]))
            r = x[idxrange]

        if return_fits is True:
            return r, disp, err, poly_disp, poly_err

    return r, disp, err


def disp2d(
    x,
    y,
    ey,
    dimy,
    smooth=True,
    bootp=True,
    a0=None,
    d0=None,
    robust_sig=False,
    nbin=None,
    nmov=None,
):
    """
    Calculates a dispersion map in the x plane.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like
        Uncertainty on the quantity from which the dispersion is calculated.
    dimy : int
        Dimension of y.
    smooth : boolean, optional
        True if the dispersion calculated from the "moving" method should
        be smoothed.
        The default is True.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is True.
    a0 : float, optional.
        Bulk RA. The default is None.
    d0 : float, optional.
        Bulk Dec. The default is None.
    robust_sig : boolean, optional
        True if the user wants to compute  a dispersion
        less sensible to outliers.
        The default is False.
    nbin : int, optional
        Auxiliar value for binning when bins is not an integer.
        The default is None.
    nmov : int, optional
        Auxiliar value for moving grids.
        The default is None.

    Returns
    -------
    r : array_like
        Reference array (possibly binned).
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    if (a0 is None) or (d0 is None):
        center, unc = position.find_center(x[0], y[0], method="iterative")
        a0 = center[0]
        d0 = center[1]

    allPoints = np.column_stack((x[0], x[1]))
    hullPoints = ConvexHull(allPoints)
    idx_lims = hullPoints.vertices
    rlims = angle.sky_distance_deg(x[0, idx_lims], x[1, idx_lims], a0, d0)
    rlim = np.nanmin(rlims)

    alim0, dlim0 = angle.get_circle_sph_trig(rlim, a0, d0)

    if nbin is None:
        nbin = int(0.25 * (position.good_bin(x[0]) + position.good_bin(x[1])))
    if nmov is None:
        nmov = int(0.7 * nbin)
    rm = rlim * 0.8

    shift = (2 * rm) / (nbin * nmov)

    amin = list()
    amax = list()

    dmin = list()
    dmax = list()

    a0 = a0 - 0.5 * nmov * shift
    d0 = d0 - 0.5 * nmov * shift

    for i in range(0, nmov):
        for j in range(0, nmov):

            alim1, dlim1 = angle.get_circle_sph_trig(rm, a0 + i * shift, d0 + j * shift)

            amin.append(np.amin(alim1))
            dmin.append(np.amin(dlim1))

            amax.append(np.amax(alim1))
            dmax.append(np.amax(dlim1))

    raux_disp = partial(aux_disp, y=y, ey=ey, dimy=dimy, robust_sig=robust_sig)
    raux_err = partial(
        aux_err, y=y, ey=ey, dimy=dimy, robust_sig=robust_sig, bootp=bootp
    )

    for i in range(0, len(amin)):

        index = np.arange(len(x[0]))

        hex_disp = plt.hexbin(
            x[0],
            x[1],
            C=index,
            gridsize=nbin,
            reduce_C_function=raux_disp,
            extent=(amin[i], amax[i], dmin[i], dmax[i]),
        )

        zax0 = hex_disp.get_array()
        verts0 = hex_disp.get_offsets()
        xax0 = np.zeros(verts0.shape[0])
        yax0 = np.zeros(verts0.shape[0])
        disp = np.zeros(verts0.shape[0])

        for offc in range(verts0.shape[0]):
            binx, biny = verts0[offc][0], verts0[offc][1]
            if zax0[offc]:
                xax0[offc], yax0[offc], disp[offc] = binx, biny, zax0[offc]

        if i == 0:
            ddisp = disp
            pointsd = np.zeros((len(xax0), 2))
            for j in range(0, len(xax0)):
                pointsd[j] = np.asarray([xax0[j], yax0[j]])
        else:
            pointsnew = np.zeros((len(xax0), 2))
            for j in range(0, len(xax0)):
                pointsnew[j] = np.asarray([xax0[j], yax0[j]])
            pointsd = np.append(pointsd, pointsnew, axis=0)
            ddisp = np.append(ddisp, disp, axis=0)

        hex_err = plt.hexbin(
            x[0],
            x[1],
            C=index,
            gridsize=nbin,
            reduce_C_function=raux_err,
            extent=(amin[i], amax[i], dmin[i], dmax[i]),
        )

        zax0 = hex_err.get_array()
        verts0 = hex_err.get_offsets()
        xax0 = np.zeros(verts0.shape[0])
        yax0 = np.zeros(verts0.shape[0])
        err = np.zeros(verts0.shape[0])

        for offc in range(verts0.shape[0]):
            binx, biny = verts0[offc][0], verts0[offc][1]
            if zax0[offc]:
                xax0[offc], yax0[offc], err[offc] = binx, biny, zax0[offc]

        if i == 0:
            eerr = err
            pointse = np.zeros((len(xax0), 2))
            for j in range(0, len(xax0)):
                pointse[j] = np.asarray([xax0[j], yax0[j]])
        else:
            pointsnew = np.zeros((len(xax0), 2))
            for j in range(0, len(xax0)):
                pointsnew[j] = np.asarray([xax0[j], yax0[j]])
            pointse = np.append(pointse, pointsnew, axis=0)
            eerr = np.append(eerr, err, axis=0)

    grid_x, grid_y = np.mgrid[
        np.amin(alim0) : np.amax(alim0) : 300j, np.amin(dlim0) : np.amax(dlim0) : 300j
    ]

    grid_disp = griddata(pointsd, ddisp, (grid_x, grid_y), method="cubic")
    grid_err = griddata(pointse, eerr, (grid_x, grid_y), method="cubic")

    grid_disp[np.isnan(grid_disp)] = 0
    grid_err[np.isnan(grid_err)] = 0

    disp = grid_disp
    err = grid_err
    r = np.asarray([pointsd, pointse])

    plt.clf()

    return r, disp, err


def dispersion(
    x,
    y,
    ey=None,
    bins=None,
    smooth=True,
    bootp=False,
    logx=False,
    nbin=None,
    polorder=None,
    return_fits=False,
    robust_sig=False,
    a0=None,
    d0=None,
    nmov=None,
):
    """
    Calculates the dispersion.

    Parameters
    ----------
    x : array_like
        Reference array.
    y : array_like
        Quantity from which the dispersion is calculated.
    ey : array_like, optional
        Uncertainty on the quantity from which the dispersion is calculated.
        Default is None.
    bins : int, string, optional
        Number of bins or method used to bin the data.
        "moving" stands for a moving grid, later interpolated with a
        cubic spline.
        The default is None.
    smooth : boolean, optional
        True if the dispersion calculated from the "moving" method should
        be smoothed.
        The default is True.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is False.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    nbin : int, optional
        Auxiliar value for binning when bins is not an integer.
        The default is None.
    polorder : int, optional
        Order of smoothing polynomial.
        The default is None.
    return_fits: Whether the user wants to return the polynomial
        smoothing fits.
        The default is False.
    robust_sig : boolean, optional
        True if the user wants to compute  a dispersion
        less sensible to outliers.
        The default is False.
    a0 : float, optional.
        Bulk RA. The default is None.
    d0 : float, optional.
        Bulk Dec. The default is None.
    nmov : int, optional
        Auxiliar value for moving grids.
        The default is None.

    Returns
    -------
    r : array_like
        Reference array (possibly binned).
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    if len(np.shape(x)) == 1:
        dimx = 1
        if bins is None:
            bins = "percentile"
    else:
        dimx = 2
        x = np.asarray(x)

    if bins is None:
        bins = "moving"

    if len(np.shape(y)) == 1:
        dimy = 1
        if ey is None:
            ey = np.asarray([np.zeros(len(y))])
        else:
            ey = np.asarray([ey])
        y = np.asarray([y])
    else:
        y = np.asarray(y)
        if ey is None:
            ey = np.zeros(np.shape(y))
        else:
            ey = np.asarray(ey)
        dimy = np.shape(y)[0]

    if dimx == 1:
        r, disp, err = disp1d(
            x,
            y,
            ey,
            dimy,
            bins=bins,
            smooth=smooth,
            bootp=bootp,
            polorder=polorder,
            logx=logx,
            nbin=nbin,
            return_fits=return_fits,
        )
    else:
        r, disp, err = disp2d(
            x,
            y,
            ey,
            dimy,
            smooth=smooth,
            bootp=bootp,
            a0=a0,
            d0=d0,
            robust_sig=robust_sig,
            nbin=nbin,
            nmov=nmov,
        )

    err[np.where(err <= 0)] = np.nan

    return r, disp, err


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"General functions"
# ------------------------------------------------------------------------------


def btsp(x):
    """
    Selects randomly a sub array of elements from x, by shuffling it.

    Parameters
    ----------
    x : array_like
        Elements to be shuffled.

    Returns
    -------
    x_new : array_like
        Shuffled array.

    """
    q = np.random.rand(len(x))
    ind = np.rint(len(x) * q - 0.5).astype(int)
    x_new = x[ind]
    return x_new


def bootstrap(array):
    """
    Bootstrap method: Computes a dispersion whitin an array of values.

    Parameters
    ----------
    array : array_like
        Array to calculate the dispersion.

    Returns
    -------
    unc : float
        Uncertainty associated to array.

    """

    if len(array) > 30:
        dof = 0
    else:
        dof = 1

    sig = np.zeros(100)
    for i in range(100):
        xb = btsp(array)
        sig[i] = np.std(xb, ddof=dof)

    unc = np.std(sig)

    return unc


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Plotting functions"
# ------------------------------------------------------------------------------


def plot_disp1D(
    ra,
    dec,
    pmra,
    epmra,
    pmdec,
    epmdec,
    corrpm,
    ra0,
    dec0,
    pmra0,
    epmra0,
    pmdec0,
    epmdec0,
    vlos0,
    d0,
    logx=False,
    logy=False,
    bootp=False,
    polorder=None,
    return_fits=False,
):
    """
    Plots a 1D dispersion map.

    Parameters
    ----------
    ra : array_like
        Gaia designation: ra.
    dec : array_like
        Gaia designation: dec.
    pmra : array_like
        Gaia designation: pmra.
    pmdec : array_like
        Gaia designation: pmdec.
    epmra : array_like
        Gaia designation: pmra_error.
    epmdec : array_like
        Gaia designation: pmdec_error.
    corrpm : array_like
        Gaia designation: pmra_pmdec_corr.
    ra0 : float
        Bulk RA, in degrees.
    dec0 : float
        Bulk Dec, in degrees.
    pmra0 : float
        Bulk PMRA, in mas/yr.
    epmra0 : float
        Bulk PMRA uncertainty, in mas/yr.
    pmdec0 : float
        Bulk PMDec, in mas/yr.
    epmdec0 : float
        Bulk PMDec uncertainty, in mas/yr.
    vlos0 : float
        Bulk line-of-sight velocity, in km/s.
    d0 : float
        Bulk distance, in kpc.
        The default is True.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is False.
    logx : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of x.
        The default is False.
    logy : boolean, optional
        True if the dispersion is evaluated in a logarithm-spaced grid of y.
        The default is False.
    polorder : int, optional
        Order of smoothing polynomial.
        The default is None.
    return_fits: Whether the user wants to return the polynomial
        smoothing fits.
        The default is False.

    Returns
    -------
    r : array_like
        Reference array (possibly binned).
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    pmr, pmt = v_sky_to_polar(ra, dec, pmra, pmdec, ra0, dec0, pmra0, pmdec0)
    rproj = angle.sky_distance_deg(ra, dec, ra0, dec0)
    pmr = pmr - pmr_corr(vlos0, rproj, d0)

    uncpmr, uncpmt = unc_sky_to_polar(
        ra, dec, epmra, epmdec, corrpm, ra0, dec0, epmra0, epmdec0
    )

    # Sorts the values according to R_proj
    L = sorted(zip(rproj, pmr, pmt, uncpmr, uncpmt), key=operator.itemgetter(0))
    rproj, pmr, pmt, uncpmr, uncpmt = zip(*L)

    rproj = np.asarray(rproj)
    pmr = np.asarray(pmr)
    pmt = np.asarray(pmt)
    uncpmr = np.asarray(uncpmr)
    uncpmt = np.asarray(uncpmt)

    rr, dd, err = dispersion(
        rproj,
        [pmr, pmt],
        [uncpmr, uncpmt],
        bins="percentile",
        polorder=polorder,
        bootp=bootp,
        logx=logx,
        return_fits=return_fits,
    )

    rr2, dd2, err2 = dispersion(
        rproj,
        [pmr, pmt],
        [uncpmr, uncpmt],
        bins=int(0.5 * position.good_bin(rproj)),
        bootp=False,
        logx=True,
        smooth=False,
    )

    if logx is False and logy is False:
        plt.plot(rr, dd, ls="-", color="red", lw=3)
        plt.plot(rr, dd + err, ls="--", color="red")
        plt.plot(rr, dd - err, ls="--", color="red")
        plt.plot(rr2, dd2, "bo")

    if logx is True and logy is False:
        plt.semilogx(rr, dd, ls="-", color="red", lw=3)
        plt.semilogx(rr, dd + err, ls="--", color="red")
        plt.semilogx(rr, dd - err, ls="--", color="red")
        plt.semilogx(rr2, dd2, "bo")

    if logx is False and logy is True:
        plt.semilogy(rr, dd, ls="-", color="red", lw=3)
        plt.semilogy(rr, dd + err, ls="--", color="red")
        plt.semilogy(rr, dd - err, ls="--", color="red")
        plt.semilogy(rr2, dd2, "bo")

    if logx is True and logy is True:
        plt.loglog(rr, dd, ls="-", color="red", lw=3)
        plt.loglog(rr, dd + err, ls="--", color="red")
        plt.loglog(rr, dd - err, ls="--", color="red")
        plt.loglog(rr2, dd2, "bo")

    plt.errorbar(rr2, dd2, yerr=err2, color="b", ls="none", barsabove=True, zorder=10)

    plt.show()

    if return_fits is True:
        return rr2, dd2, err2
    else:
        return


def plot_disp2D(
    ra,
    dec,
    pmra,
    epmra,
    pmdec,
    epmdec,
    corrpm,
    ra0,
    dec0,
    pmra0,
    epmra0,
    pmdec0,
    epmdec0,
    vlos0,
    d0,
    rlim=None,
    bootp=False,
    robust_sig=False,
    nbin=None,
    nmov=None,
    return_fits=False,
):
    """
    Plots a 2D dispersion map, along with a map of respective uncertainties.

    Parameters
    ----------
    ra : array_like
        Gaia designation: ra.
    dec : array_like
        Gaia designation: dec.
    pmra : array_like
        Gaia designation: pmra.
    pmdec : array_like
        Gaia designation: pmdec.
    epmra : array_like
        Gaia designation: pmra_error.
    epmdec : array_like
        Gaia designation: pmdec_error.
    corrpm : array_like
        Gaia designation: pmra_pmdec_corr.
    ra0 : float
        Bulk RA, in degrees.
    dec0 : float
        Bulk Dec, in degrees.
    pmra0 : float
        Bulk PMRA, in mas/yr.
    epmra0 : float
        Bulk PMRA uncertainty, in mas/yr.
    pmdec0 : float
        Bulk PMDec, in mas/yr.
    epmdec0 : float
        Bulk PMDec uncertainty, in mas/yr.
    vlos0 : float
        Bulk line-of-sight velocity, in km/s.
    d0 : float
        Bulk distance, in kpc.
    rlim : float, optional
        Maximum radius in the image, in degrees.
        The default is None.
    bootp : boolean, optional
        True if the errors are drawn from a Bootstrap method.
        The default is False.
    robust_sig : boolean, optional
        True if the user wants to compute  a dispersion
        less sensible to outliers.
        The default is False.
    nbin : int, optional
        Auxiliar value for binning when bins is not an integer.
        The default is None.
    nmov : int, optional
        Auxiliar value for moving grids.
        The default is None.
    return_fits : boolean, optional
        True is user wants to return the dispersion profile.
        The default is False.

    Returns
    -------
    r : array_like
        Reference array (possibly binned).
    disp : array_like
        Dispersion.
    err : array_like
        Uncertainty on the dispersion.

    """

    pmr, pmt = v_sky_to_polar(ra, dec, pmra, pmdec, ra0, dec0, pmra0, pmdec0)
    rproj = angle.sky_distance_deg(ra, dec, ra0, dec0)
    pmr = pmr - pmr_corr(vlos0, rproj, d0)

    uncpmr, uncpmt = unc_sky_to_polar(
        ra, dec, epmra, epmdec, corrpm, ra0, dec0, epmra0, epmdec0
    )

    rr, dd, err = dispersion(
        [ra, dec],
        [pmr, pmt],
        [uncpmr, uncpmt],
        bootp=bootp,
        a0=ra0,
        d0=dec0,
        robust_sig=robust_sig,
        nbin=nbin,
        nmov=nmov,
    )

    if rlim is None:
        rlim = position.quantile(rproj, 0.5)
    xlim, ylim = angle.get_circle_sph_trig(rlim, ra0, dec0)

    allPoints = np.column_stack((ra, dec))
    hullPoints = ConvexHull(allPoints)
    idx_lims = hullPoints.vertices
    rlims = angle.sky_distance_deg(ra[idx_lims], dec[idx_lims], ra0, dec0)
    rmax = np.nanmin(rlims)

    alim0, dlim0 = angle.get_circle_sph_trig(rmax, ra0, dec0)

    qmin = position.quantile(dd.flatten()[np.where(dd.flatten() > 0)], 0.05)
    qmax = position.quantile(dd.flatten()[np.where(dd.flatten() > 0)], 0.95)

    fig, ax = plt.subplots(figsize=(5, 4))
    plt.title(r"Dispersion")
    c = plt.imshow(
        dd.T,
        origin="lower",
        aspect="auto",
        vmin=qmin,
        vmax=qmax,
        cmap="jet",
        extent=[np.nanmin(alim0), np.nanmax(alim0), np.nanmin(dlim0), np.nanmax(dlim0)],
    )
    cbar = fig.colorbar(c)
    cbar.ax.tick_params(labelsize=13)
    plt.xlim([np.amin(xlim), np.amax(xlim)])
    plt.ylim([np.amin(ylim), np.amax(ylim)])
    plt.gca().invert_xaxis()
    plt.show()

    qmin = position.quantile(err.flatten()[np.where(err.flatten() > 0)], 0.05)
    qmax = position.quantile(err.flatten()[np.where(err.flatten() > 0)], 0.95)

    fig, ax = plt.subplots(figsize=(5, 4))
    plt.title(r"Uncertainty in dispersion")
    c = plt.imshow(
        err.T,
        origin="lower",
        aspect="auto",
        vmin=qmin,
        vmax=qmax,
        cmap="jet",
        extent=[np.nanmin(alim0), np.nanmax(alim0), np.nanmin(dlim0), np.nanmax(dlim0)],
    )
    cbar = fig.colorbar(c)
    cbar.ax.tick_params(labelsize=13)
    plt.xlim([np.amin(xlim), np.amax(xlim)])
    plt.ylim([np.amin(ylim), np.amax(ylim)])
    plt.gca().invert_xaxis()
    plt.show()

    if return_fits is True:
        return rr, dd, err
    else:
        return



"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning the color magnitude
# diagram (CMD).
# It provides a Kernel Density Estimation (KDE) of the CMD distribution.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

from scipy.stats import gaussian_kde
from shapely.geometry import Point
from shapely.geometry.polygon import Polygon
import numpy as np

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"KDE approach"
# ------------------------------------------------------------------------------


def kde(x, y, ww=None, bw="silverman", confidence=None):
    """
    This function computes the KDE estimative of the PDF of the color
    magnitude diagram (CMD).

    Parameters
    ----------
    x : array_like
        Color from the CMD.
    y : array_like
        Magnitude.
    ww : array_like, optional
        Weight passed to the KDE estimation. The default is None.
    bw : string or float, optional
        Argument "bw_method" from the "gaussian_kde" method from scipy.stats.
        The default is 'silverman'.
    confidence : float, optional
        Confidence level (in percent) of arrays x and y to be considered.
        Default is None.

    Returns
    -------
    Z : array of size [2, Nsamples]
        KDE estimative of the PDF.

    """

    if ww is None:
        ww = np.ones(len(x))

    if confidence is not None:
        conf = confidence / 100
        qx_i, qx_f = quantile(x, [conf, 1 - conf])
        qy_i, qy_f = quantile(y, [conf, 1 - conf])
    else:
        qx_i, qx_f = np.nanmin(x), np.nanmax(x)
        qy_i, qy_f = np.nanmin(y), np.nanmax(y)

    ranges = [[qx_i, qx_f], [qy_i, qy_f]]

    # fit an array of size [Ndim, Nsamples]
    data = np.vstack([x, y])
    kernel = gaussian_kde(data, bw_method=bw, weights=ww)

    q_16, q_50, q_84 = quantile(x, [0.16, 0.5, 0.84])
    q_m, q_p = q_50 - q_16, q_84 - q_50
    nbins_x = int(10 * (np.amax(x) - np.amin(x)) / (min(q_m, q_p)))

    q_16, q_50, q_84 = quantile(y, [0.16, 0.5, 0.84])
    q_m, q_p = q_50 - q_16, q_84 - q_50
    nbins_y = int(10 * (np.amax(y) - np.amin(y)) / (min(q_m, q_p)))

    # evaluate on a regular grid
    xgrid = np.linspace(ranges[0][0], ranges[0][1], nbins_x)
    ygrid = np.linspace(ranges[1][0], ranges[1][1], nbins_y)
    Xgrid, Ygrid = np.meshgrid(xgrid, ygrid)
    Zhist = kernel.evaluate(np.vstack([Xgrid.ravel(), Ygrid.ravel()]))
    Z = Zhist.reshape(Xgrid.shape)

    return Z


def bw_silver(x, y):
    """
    Returns the Silverman bandwidth factor.

    Parameters
    ----------
    x : array_like
        Color from the CMD.
    y : array_like
        Magnitude.

    Returns
    -------
    bw : float
        Silverman's bandwidth factor.

    """

    d = 2
    n = len(x)

    bw = (n * (d + 2) / 4) ** (-1 / (d + 4))

    return bw


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"General functions"
# ---------------------------------------------------------------------------


def quantile(x, q):
    """
    Compute sample quantiles.

    Parameters
    ----------
    x : array_like
        Array containing set of values.
    q : array_like, float
        Quantile values to be derived (must be between 0 and 1).

    Raises
    ------
    ValueError
        Quantiles must be between 0 and 1.

    Returns
    -------
    percentile: array_like (same shape as q), float
        Percentile value(s).

    """

    x = np.atleast_1d(x)
    q = np.atleast_1d(q)

    if np.any(q < 0.0) or np.any(q > 1.0):
        raise ValueError("Quantiles must be between 0 and 1")

    return np.percentile(x, 100.0 * q)


def inside_contour(x, y, contours, idx=None, idx_level=0):
    """
    Returns indexes from a subset of previous indexes corresponding
    to the points in (x,y) inside a given contour.

    Parameters
    ----------
    x : array_like
        Data in x-direction.
    y : array_like
        Data in y-direction.
    contours : QuadContourSet
        Contour object from pyplot.
    idx : array of integers, optional
        Array containing the indexes in (x,y) from which to search points
        inside a given contour. The default is None.
    idx_level : int, optional
        Which contour lines to consider from contour object.
        The default is 0 (wider contour).

    Returns
    -------
    idx_inside : array of integers
        Array of indexes inside the contour.

    """

    lcontours = list()
    inside = list()

    p = contours.collections[0].get_paths()
    npoly = len(p)
    for i in range(0, npoly):
        v = p[i].vertices
        contourx = v[:, 0]
        contoury = v[:, 1]

        contour_array = np.array([contourx, contoury])

        lcontours.append(contour_array)
        inside.append(
            Polygon([(contourx[i], contoury[i]) for i in range(0, len(contourx))])
        )

    idx_inside = list()
    for i in idx:
        for j in range(0, npoly):
            if inside[j].contains(Point(x[i], y[i])):
                idx_inside.append(i)

    idx_inside = np.asarray(idx_inside).astype(int)

    return idx_inside



"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning the angular tansformations,
# sky projections and spherical trigonomtry.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

import numpy as np

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Angles handling"
# ------------------------------------------------------------------------------


def sky_distance_deg(RA, Dec, RA0, Dec0):
    """
    Computes the sky distance (in degrees) between two sets of
    sky coordinates, given also in degrees.

    Parameters
    ----------
    RA : array_like, float
        Right ascension (in degrees) of object 1.
    Dec : array_like (same shape as RA), float
        Declination (in degrees) of object 1.
    RA0 : array_like (same shape as RA), float
        Right ascension (in degrees) of object 2.
    Dec0 : array_like (same shape as RA), float
        Declination (in degrees) of object 2.

    Returns
    -------
    R : array_like, float
        Sky distance (in degrees) between object 1 and object 2.

    """

    RA = RA * np.pi / 180
    Dec = Dec * np.pi / 180

    RA0 = RA0 * np.pi / 180
    Dec0 = Dec0 * np.pi / 180

    R = (180 / np.pi) * np.arccos(
        np.sin(Dec) * np.sin(Dec0) + np.cos(Dec) * np.cos(Dec0) * np.cos((RA - RA0))
    )

    return np.asarray(R)


def get_circle_sph_trig(r, a0, d0, nbins=500):
    """
    Generates a circle in spherical coordinates.

    Parameters
    ----------
    r : float
        Distance from the center, in degrees.
    a0 : float
        Right ascention from origin, in degrees.
    d0 : float
        Declination from origin in, in degrees.
    nbins : int, optional
        Number of circle points. The default is 500.

    Returns
    -------
    ra : array_like
        Right ascention in degrees.
    dec : array_like
        Declination in degrees.

    """

    # Converts angles to radians
    r = r * np.pi / 180
    a0 = a0 * np.pi / 180
    d0 = d0 * np.pi / 180

    phi = np.linspace(0, 2 * np.pi, nbins)

    a = np.zeros(nbins)
    d = np.zeros(nbins)
    for i in range(0, nbins):
        a[i], d[i] = polar_to_sky(r, phi[i], a0, d0)

    return a, d


def polar_to_sky(r, phi, a0, d0):
    """
    Transforms spherical polar coordinates (r,phi) into sky coordinates,
    in degrees (RA,Dec).

    Parameters
    ----------
    r : array_like
        Radial distance from center.
    phi : array_like
        Angle between increasing declination and the projected radius
        (pointing towards the source).
    a0 : float
        Right ascention from origin in radians.
    d0 : float
        Declination from origin in radians.

    Returns
    -------
    ra : array_like
        Right ascention in degrees.
    dec : array_like
        Declination in degrees.

    """

    d = np.arcsin(np.cos(r) * np.sin(d0) + np.cos(d0) * np.cos(phi) * np.sin(r))

    if phi < np.pi:
        if (np.cos(r) - np.sin(d) * np.sin(d0)) / (np.cos(d) * np.cos(d0)) > 0:
            a = a0 + np.arccos(np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))
        else:
            a = a0 + np.arccos(-np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))

    if phi >= np.pi:
        if (np.cos(r) - np.sin(d) * np.sin(d0)) / (np.cos(d) * np.cos(d0)) > 0:
            a = a0 - np.arccos(np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))
        else:
            a = a0 - np.arccos(-np.sqrt(1 - (np.sin(phi) * np.sin(r) / np.cos(d)) ** 2))

    ra = a * 180 / np.pi
    dec = d * 180 / np.pi

    return ra, dec


def sky_to_polar(a, d, a0, d0):
    """
    Transforms sky coordinates, in degrees (RA,Dec), into
    spherical polar coordinates (r,phi).

    Parameters
    ----------
    a : array_like
        Right ascention in degrees.
    d : array_like
        Declination in degrees.
    a0 : float
        Right ascention from origin.
    d0 : float
        Declination from origin.

    Returns
    -------
    r : array_like
        Radial distance from center.
    p : array_like
        Angle between increasing declination and the projected radius
        (pointing towards the source), in radians.

    """

    r = sky_distance_deg(a, d, a0, d0) * np.pi / 180

    a = a * np.pi / 180
    d = d * np.pi / 180

    sp = np.cos(d) * np.sin(a - (a0 * np.pi / 180)) / np.sin(r)

    p = np.zeros(len(sp))

    spp = np.where(sp > 0)
    spm = np.where(sp <= 0)
    dp = np.where(d > (d0 * np.pi / 180))
    dm = np.where(d <= (d0 * np.pi / 180))

    p[np.intersect1d(spp, dp)] = np.arcsin(sp[np.intersect1d(spp, dp)])
    p[np.intersect1d(spp, dm)] = np.pi - np.arcsin(sp[np.intersect1d(spp, dm)])
    p[np.intersect1d(spm, dp)] = 2 * np.pi + np.arcsin(sp[np.intersect1d(spm, dp)])
    p[np.intersect1d(spm, dm)] = np.pi - np.arcsin(sp[np.intersect1d(spm, dm)])

    return r, p


def angular_sep_vector(v0, v):
    """
    Returns separation angle in radians between two 3D arrays.

    Parameters
    ----------
    v0 : 3D array
        Vector 1.
    v : 3D array
        Vector 2.

    Returns
    -------
    R : array_like
        Separation between vector 1 and vector 2 in radians.

    """

    try:
        v0.shape
    except NameError:
        print("You did not give a valid input.")
        return
    v = v / np.linalg.norm(v)
    B = np.arcsin(v[2])
    cosA = v[0] / np.cos(B)
    sinA = v[1] / np.cos(B)

    if cosA > 0 and sinA >= 0:
        A = np.arcsin(sinA)
    if cosA <= 0 and sinA > 0:
        A = np.arccos(cosA)
    if cosA < 0 and sinA <= 0:
        A = 2 * np.pi - np.arccos(cosA)
    if cosA >= 0 and sinA < 0:
        A = np.arcsin(sinA)

    v0 = v0 / np.linalg.norm(v0)
    B0 = np.arcsin(v0[2])
    cosA0 = v0[0] / np.cos(B0)
    sinA0 = v0[1] / np.cos(B0)

    if cosA0 > 0 and sinA0 >= 0:
        A0 = np.arcsin(sinA0)
    if cosA0 <= 0 and sinA0 > 0:
        A0 = np.arccos(cosA0)
    if cosA0 < 0 and sinA0 <= 0:
        A0 = 2 * np.pi - np.arccos(cosA0)
    if cosA0 >= 0 and sinA0 < 0:
        A0 = np.arcsin(sinA0)

    cosR = np.sin(B0) * np.sin(B) + np.cos(B0) * np.cos(B) * np.cos(A - A0)
    R = np.arccos(cosR)

    return R


def rodrigues_formula(k, v, theta, debug=False):
    """
    Returns the rotation of v of an angle theta with respect to the vector k
    Applies the Rodrigues formula from:
    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

    Parameters
    ----------
    k : array_like
        Vector with respect to which v will be rotated.
    v : array_like
        Vector to be rotated.
    theta : float
        Angle to rotate v.
    debug : boolean, optional
       True if the reader wants to print debug diagnistics.
       The default is False.

    Returns
    -------
    v_rot : array_like
        Rotated vector.

    """

    try:
        v.shape
    except NameError:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    if len(np.shape(v)) == 1 and len(v) == 3:
        v_rot = (
            v * np.cos(theta)
            + np.cross(k, v) * np.sin(theta)
            + k * np.dot(k, v) * (1 - np.cos(theta))
        )

    elif len(np.shape(v)) == 2 and np.shape(v)[0] == 3:
        v_rot = np.zeros((np.shape(v)[1], 3))
        for i in range(0, len(v_rot)):

            v0 = np.asarray([v[0][i], v[1][i], v[2][i]])
            v_rot[i] = (
                v0 * np.cos(theta)
                + np.cross(k, v0) * np.sin(theta)
                + k * np.dot(v0, k) * (1 - np.cos(theta))
            )

            if debug is True and i < 10:
                print("v0   :", v0)
                print("v_rot:", v_rot[i])
    else:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    return v_rot


def sky_coord_rotate(v_i, v0_i, v0_f, theta=0, debug=False):
    """
    Gets new angles (RA,Dec) in degrees of a rotated vector.

    Parameters
    ----------
    v_i : array_like
        Vectors to be rotated.
    v0_i : array_like
        Vector pointing to the initial centroid position.
    v0_f : array_like
        Vector pointing to the final centroid position.
    theta : float, optional
        Angle to rotate (for no translation), in radians. The default is 0.
    debug : boolean, optional
       True if the reader wants to print debug diagnistics.
       The default is False.

    Returns
    -------
    array_like, array_like
        Arrays containing the new rotated (RA,Dec) positions.

    """

    if (v0_i == v0_f).all():
        if debug is True:
            print("Pure rotation")
        k = v0_i / np.linalg.norm(v0_i)
    else:
        if debug is True:
            print("Translation in spherical geometry")
        k = np.cross(v0_i, v0_f) / np.linalg.norm(np.cross(v0_i, v0_f))
        theta = angular_sep_vector(v0_i, v0_f)

    if debug is True:
        print("Vector k:", k)
        print("Angle of separation [degrees]:", theta * 180 / np.pi)

    v_f = rodrigues_formula(k, v_i, theta)

    try:
        v_f.shape
    except NameError:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    if len(np.shape(v_f)) == 1 and len(v_f) == 3:
        v_f = v_f / np.linalg.norm(v_f)
        B = np.arcsin(v_f[2])
        cosA = v_f[0] / np.cos(B)
        sinA = v_f[1] / np.cos(B)

        if cosA > 0 and sinA >= 0:
            A = np.arcsin(sinA)
        if cosA <= 0 and sinA > 0:
            A = np.arccos(cosA)
        if cosA < 0 and sinA <= 0:
            A = 2 * np.pi - np.arccos(cosA)
        if cosA >= 0 and sinA < 0:
            A = np.arcsin(sinA)

    elif len(np.shape(v_f)) == 2:
        A = np.zeros(len(v_f))
        B = np.zeros(len(v_f))
        for i in range(0, len(v_f)):
            v_f[i] = v_f[i] / np.linalg.norm(v_f[i])
            B[i] = np.arcsin(v_f[i][2])
            cosA = v_f[i][0] / np.cos(B[i])
            sinA = v_f[i][1] / np.cos(B[i])

            if cosA > 0 and sinA >= 0:
                A[i] = np.arcsin(sinA)
            if cosA <= 0 and sinA > 0:
                A[i] = np.arccos(cosA)
            if cosA < 0 and sinA <= 0:
                A[i] = 2 * np.pi - np.arccos(cosA)
            if cosA >= 0 and sinA < 0:
                A[i] = np.arcsin(sinA)

            if debug is True and i < 10:
                print("A, B [degrees]:", A[i] * (180 / np.pi), B[i] * (180 / np.pi))
    else:
        print("You did not give a valid input for the Rodrigues formula.")
        return

    return A * (180 / np.pi), B * (180 / np.pi)


def sky_vector(a, d, a0, d0, af, df):
    """
    Transforms sky coordinates in vectors to be used by sky_coord_rotate.

    Parameters
    ----------
    a : float, array_like
        Original set of RA in degrees.
    d : float, array_like
        Original set of Dec in degrees.
    a0 : float
        Original centroid RA in degrees.
    d0 : float
        Original centroid Dec in degrees.
    af : float
        Final centroid RA in degrees.
    df : float
        Final centroid Dec in degrees.

    Returns
    -------
    v_i : array_like
        Vectors to be rotated.
    v0_i : array_like
        Vector pointing to the initial centroid position.
    v0_f : array_like
        Vector pointing to the final centroid position.

    """

    a = a * (np.pi / 180)
    d = d * (np.pi / 180)
    a0 = a0 * (np.pi / 180)
    d0 = d0 * (np.pi / 180)
    af = af * (np.pi / 180)
    df = df * (np.pi / 180)

    v_i = np.asarray([np.cos(a) * np.cos(d), np.sin(a) * np.cos(d), np.sin(d)])

    v0_i = np.asarray([np.cos(a0) * np.cos(d0), np.sin(a0) * np.cos(d0), np.sin(d0)])

    v0_f = np.asarray([np.cos(af) * np.cos(df), np.sin(af) * np.cos(df), np.sin(df)])

    return v_i, v0_i, v0_f


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Axis rotation"
# ------------------------------------------------------------------------------


def transrot_source(a, d, a0, d0, af, df):
    """
    Translades and then rotates a source in spherical coordinates, so
    the original directions remain the same.

    Parameters
    ----------
    a : float, array_like
        Original set of RA in degrees.
    d : float, array_like
        Original set of Dec in degrees.
    a0 : float
        Original centroid RA in degrees.
    d0 : float
        Original centroid Dec in degrees.
    af : float
        Final centroid RA in degrees.
    df : float
        Final centroid Dec in degrees.

    Returns
    -------
    a : array_like
        Right ascention in degrees.
    d : array_like
        Declination in degrees.

    """

    v_i, v0_i, v0_f = sky_vector(a, d, a0, d0, af, df)

    a, d = sky_coord_rotate(v_i, v0_i, v0_f)

    rmax = np.nanmax(sky_distance_deg(a, d, af, df)) * (np.pi / 180)

    narr = np.asarray([0, 0.1, 0.25, 0.5]) * rmax
    vt = np.asarray(
        [
            np.cos(a0 + narr) * np.cos([d0, d0, d0, d0]),
            np.sin(a0 + narr) * np.cos([d0, d0, d0, d0]),
            np.sin([d0, d0, d0, d0]),
        ]
    )

    at, dt = sky_coord_rotate(vt, v0_i, v0_f)

    r0, p0 = sky_to_polar(a0 + narr, d0 * np.ones(len(narr)), a0, d0)
    rf, pf = sky_to_polar(at, dt, af, df)

    phi = np.nanmean(pf - p0)

    v_i, v0_i, v0_f = sky_vector(a, d, af, df, af, df)

    a, d = sky_coord_rotate(v_i, v0_i, v0_f, theta=phi)

    return a, d


def rotate_axis(x, y, theta, mu_x=0, mu_y=0):
    """
    Rotates and translates the two main cartesian axis.

    Parameters
    ----------
    x : float or array_like
        Data in x-direction.
    y : float or array_like
        Data in y-direction.
    theta : float
        Rotation angle in radians.
    mu_x : float, optional
        Center of new x axis in the old frame. The default is 0.
    mu_y : float, optional
        Center of new y axis in the old frame. The default is 0.

    Returns
    -------
    x_new : float or array_like
        Data in new x-direction.
    y_new : float or array_like
        Data in new y-direction.

    """

    x_new = (x - mu_x) * np.cos(theta) + (y - mu_y) * np.sin(theta)
    y_new = -(x - mu_x) * np.sin(theta) + (y - mu_y) * np.cos(theta)

    return x_new, y_new


def get_ellipse(a, b, theta, nbins):
    """
    Provides an array describing a rotated ellipse.

    Parameters
    ----------
    a : float
        Ellipse semi-major axis.
    b : float
        Ellipse semi-minor axis.
    theta : float
        Rotation angle in radians.
    nbins : int
        Number of points in the final array.

    Returns
    -------
    ellipse_rot : array_like
        Rotated ellipse array.

    """

    t = np.linspace(0, 2 * np.pi, nbins)
    ellipse = np.array([a * np.cos(t), b * np.sin(t)])
    m_rot = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    ellipse_rot = np.zeros((2, ellipse.shape[1]))
    for i in range(ellipse.shape[1]):
        ellipse_rot[:, i] = np.dot(m_rot, ellipse[:, i])

    return ellipse_rot


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ------------------------------------------------------------------------------
"Transform coordinates"
# ------------------------------------------------------------------------------


def cart_to_sph(x, y, z, vx, vy, vz):
    """
    Transforms 6D cartesian coordinates to spherical coordinates.

    Parameters
    ----------
    x : array_like
        x-axis.
    y : array_like
        y-axis.
    z : array_like
        z-axis.
    vx : array_like
        x-axis velocity.
    vy : array_like
        y-axis velocity.
    vz : array_like
        z-axis velocity.

    Returns
    -------
    r : array_like
        r-axis.
    phi : array_like
        phi-angle.
    theta : array_like
        theta-angle.
    vr : array_like
        r-axis velocity.
    vphi : array_like
        phi-angle velocity.
    vtheta : array_like
        theta-angle velocity.

    """

    r = np.sqrt(x * x + y * y + z * z)
    phi = np.arctan(y / x)
    theta = np.arccos(z / r)

    vr = (vx * x + vy * y + vz * z) / r
    vphi = r * (vy * x - vx * y) / (x * x + y * y)
    vtheta = (vz * (x * x + y * y) - z * (vx * x + vy * y)) / (
        np.sqrt(x * x + y * y) * r
    )

    return r, phi, theta, vr, vphi, vtheta


def sph_to_cart(r, phi, theta, vr, vphi, vtheta):
    """
    Transforms 6D spherical coordinates to cartesian coordinates.

    Parameters
    ----------
    r : array_like
        r-axis.
    phi : array_like
        phi-angle.
    theta : array_like
        theta-angle.
    vr : array_like
        r-axis velocity.
    vphi : array_like
        phi-angle velocity.
    vtheta : array_like
        theta-angle velocity.

    Returns
    -------
    x : array_like
        x-axis.
    y : array_like
        y-axis.
    z : array_like
        z-axis.
    vx : array_like
        x-axis velocity.
    vy : array_like
        y-axis velocity.
    vz : array_like
        z-axis velocity.

    """

    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)

    vx = (
        vr * np.sin(theta) * np.cos(phi)
        + vtheta * np.cos(theta) * np.cos(phi)
        - vphi * np.sin(theta) * np.sin(phi)
    )
    vy = (
        vr * np.sin(theta) * np.sin(phi)
        + vtheta * np.cos(theta) * np.sin(phi)
        + vphi * np.sin(theta) * np.cos(phi)
    )
    vz = vr * np.cos(theta) - vtheta * np.sin(theta)

    return x, y, z, vx, vy, vz



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['balrogo']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.0,<5.0',
 'emcee>=3.0.2,<4.0.0',
 'matplotlib>=3.1.2,<4.0.0',
 'numdifftools>=0.9.39,<0.10.0',
 'numpy>=1.18.5,<2.0.0',
 'scikit-image>=0.16.2,<0.17.0',
 'scipy>=1.4.1,<2.0.0',
 'shapely>=1.6.4,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['importlib_metadata>=3.3.0,<4.0.0']}

setup_kwargs = {
    'name': 'balrogo',
    'version': '1.3.2',
    'description': 'Bayesian Astrometric Likelihood Recovery of Galactic Objects',
    'long_description': '# BALRoGO\n\n[![pipeline status](https://gitlab.com/eduardo-vitral/balrogo/badges/master/pipeline.svg)](https://gitlab.com/eduardo-vitral/balrogo/-/commits/master)\n[![coverage report](https://gitlab.com/eduardo-vitral/balrogo/badges/master/coverage.svg)](https://gitlab.com/eduardo-vitral/balrogo/-/commits/master)\n[![pypi](https://img.shields.io/pypi/v/balrogo.svg)](https://pypi.python.org/pypi/balrogo/)\n[![python](https://img.shields.io/pypi/pyversions/balrogo.svg)](https://pypi.python.org/pypi/balrogo)\n[![license](http://img.shields.io/badge/license-MIT-blue.svg)](LICENSE)\n\n<!-- markdownlint-disable-next-line no-inline-html -->\n<img alt="logo" align="right" src="https://gitlab.com/eduardo-vitral/balrogo/-/raw/master/logo.png" width="20%" />\n\nBALRoGO: Bayesian Astrometric Likelihood Recovery of Galactic Objects.\n\n- Specially developed to handle data from the Gaia space mission.\n- Extracts galactic objects such as globular clusters and dwarf galaxies from data contiminated by interlopers.\n- Uses a combination of Bayesian and non-Bayesian approaches.\n- Provides:\n  - Fits of proper motion space.\n  - Fits of surface density.\n  - Fits of object center.\n  - Confidence regions for the color-magnitude diagram and parallaxes.\n\nIf something does not work, please [file an issue](https://gitlab.com/eduardo-vitral/balrogo/-/issues).\n\n## Attribution\n\nPlease cite [us](https://academic.oup.com/mnras/article-abstract/504/1/1355/6237524?redirectedFrom=fulltext) if you find this code useful in your research and add your paper to the testimonials list. The BibTeX entry for the paper is:\n\n```bibtex\n@ARTICLE{Vitral21,\n   author = {{Vitral}, Eduardo},\n    title = "BALRoGO: Bayesian Astrometric Likelihood Recovery of Galactic Objects - Global properties of over one hundred globular clusters with Gaia EDR3",\n  journal = {\\mnras},\n     year = 2021,\n    month = jun,\n   volume = {504},\n   number = {1},\n    pages = {1355-1369},\n      doi = {10.1093/mnras/stab947},\n   eprint = {2102.04841},\n   adsurl = {https://ui.adsabs.harvard.edu/abs/2021MNRAS.504.1355V},\n}\n```\n\n## Quick overview\n\nBALRoGO has eight modules that perform different tasks:\n\n- ***angle.py*** : This module contains the main functions concerning angular tansformations, sky projections and spherical trigonomtry.\n- ***gaia.py*** : This module contains the main functions concerning the handling of the Gaia mission data.\n- ***hrd.py*** : This module contains the main functions concerning the color magnitude diagram (CMD). It provides a Kernel Density Estimation (KDE) of the CMD distribution.\n- ***marginals.py*** : This module is based on the Python corner package (Copyright 2013-2016 Dan Foreman-Mackey & contributors, The Journal of Open Source Software): https://joss.theoj.org/papers/10.21105/joss.00024\nI have done some modifications on it so it allows some new features and so it takes into account some choices as default. I thank Gary Mamon for his good suggestions concerning the plot visualization.\n-  ***parallax.py*** : This module contains the main functions concerning parallax information. It provides a kernel density estimation of the distance distribution, as well as a fit of the mode of this distribution.\n- ***pm.py*** : This module contains the main functions concerning proper motion data. It provides MCMC and maximum likelihood fits of proper motions data, as well as robust initial guesses for those fits.\n- ***position.py*** : This module contains the main functions concerning positional information. It provides MCMC and maximum likelihood fits of surface density, as well as robust initial guesses for the (RA,Dec) center of the source.\n- ***mock.py*** : This files handles mock data sets. It converts 3D coordinates to sky coordinates and is able to add realistic errors to proper motions. It is also able to generate Milky Way interlopers.\n\n## Installation\n\nBALRoGO is available through [pip](https://pypi.org/project/balrogo/). The quickiest way to install it is to type the following command in your terminal:\n\n```terminal\npip install balrogo\n```\n\nIf you are using [Anaconda](https://www.anaconda.com/), you might want to install it directly in your Anaconda bin path:\n\n```terminal\ncd path/anaconda3/bin/\npip install balrogo\n```\n\nFor updated versions of the code, you can do the same as above, but instead of using `pip install balrogo`, you should type:\n\n```terminal\npip install --upgrade balrogo\n```\n\n### Using BALRoGO on [*Gaia*](https://www.cosmos.esa.int/web/gaia/data-access) data\n\nFor quick tutorial of BALRoGO applied to *Gaia* data, please click [here](https://gitlab.com/eduardo-vitral/balrogo/-/blob/master/GAIA.md).\n\n## License\n\nCopyright (c) 2020 Eduardo Vitral & Alexandre Macedo.\n\nBALRoGO is free software made available under the [MIT License](LICENSE). The BALRoGO logo is licensed under a [Creative Commons Attribution 4.0 International license](https://creativecommons.org/licenses/by/4.0/).\n',
    'author': 'Eduardo Vitral',
    'author_email': 'vitral@iap.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/eduardo-vitral/balrogo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Created on 2020

@author: Eduardo Vitral
"""

###############################################################################
#
# November 2020, Paris
#
# This file contains the main functions concerning the handling of the
# Gaia mission data.
#
# Documentation is provided on Vitral, 2021.
# If you have any further questions please email vitral@iap.fr
#
###############################################################################

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mc
from astropy.io import fits
from . import pm
from . import angle
from . import position
from . import hrd

# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Gaia handling"
# ---------------------------------------------------------------------------


def polyRiello20(x):
    """
    Polynomial from Riello et al. 2020. With coefficients from their Table 2.

    Parameters
    ----------
    x : array_like
        x = G_BP − G_RP.

    Returns
    -------
    f : array_like
        BP and RP flux excess factor C dependence on the x = G_BP − G_RP.

    """

    f = np.zeros(len(x))

    for i in range(len(x)):

        if np.isnan(x[i]):
            f[i] = np.nan

        elif x[i] < 0.5:
            a0 = 1.154360
            a1 = 0.033772
            a2 = 0.032277

            f[i] = a0 + a1 * x[i] + a2 * x[i] * x[i]

        elif x[i] >= 0.5 and x[i] < 4.0:
            a0 = 1.162004
            a1 = 0.011464
            a2 = 0.049255
            a3 = -0.005879

            f[i] = a0 + a1 * x[i] + a2 * x[i] * x[i] + a3 * x[i] * x[i] * x[i]

        elif x[i] >= 4.0:
            a0 = 1.057572
            a1 = 0.140537

            f[i] = a0 + a1 * x[i]

    return f


def clean_gaia(
    pmra,
    pmdec,
    epmra,
    epmdec,
    corrpm,
    chi2,
    nu,
    g_mag,
    br_mag,
    br_excess,
    ra,
    dec,
    err_lim=None,
    err_handle="relative",
    object_type="gc",
    r_cut=None,
    c=None,
    cleaning="v21",
):
    """
    Cleans the Gaia data based on astrometric flags.

    Parameters
    ----------
    pmra : array_like
        Gaia designation: pmra.
    pmdec : array_like
        Gaia designation: pmdec.
    epmra : array_like
        Gaia designation: pmra_error.
    epmdec : array_like
        Gaia designation: pmdec_error.
    corrpm : array_like
        Gaia designation: pmra_pmdec_corr.
    chi2 : array_like
        Gaia designation: astrometric_chi2_al.
    nu : array_like
        Gaia designation: astrometric_n_good_obs_al.
    g_mag : array_like
        Gaia designation: phot_g_mean_mag.
    br_mag : array_like
        Gaia designation: bp_rp.
    br_excess : array_like
        Gaia designation: phot_bp_rp_excess_factor.
    ra : array_like
        Gaia designation: ra.
    dec : array_like
        Gaia designation: dec.
    err_lim : float, optional
        Error threshold for proper motions. The default is None.
    err_handle : string, optional
        Way in which the error limit is handled:
            - 'relative', if the err_lim argument is the number of times
            the initial guess on the galactic object proper motion dispersion.
            - 'absolute', if the err_lim argument is the absolute error
            limit in mas/yr.
        The default is 'relative'.
    object_type : string, optional
        Galactic object type, to be considered when defining the default
        value of err_lim:
            - 'gc', for globular cluster.
            - 'dsph', for dwarf spheroidal galaxy.
        The default is 'gc'.
    r_cut : float, optional
        Projected radius limit. The default is None.
    c : 2D array, optional
        (ra,dec) of the center. The default is None.
    cleaning : string, optional
        Cleaning procedure to be adopted:
            - 'v21', to use the same cleaning from Vitral 2021
            - 'vm21', to use the same cleaning from Vitral & Mamon 21
        The default is 'v21'.

    Raises
    ------
    ValueError
        err_handle is not one of the following:
            - 'relative'
            - 'absolute'
        object_type is not one of the following:
            - 'gc'
            - 'dsph'
        cleaning is not one of the following:
            - 'v21'
            - 'vm21'

    Returns
    -------
    idx : array_like
        Array containing the indexes of the original provided arrays
        which correspond to stars with clean astrometry.

    """

    if err_handle not in ["relative", "absolute"]:
        raise ValueError("Does not recognize method to handle errors.")

    if object_type not in ["gc", "dsph"]:
        raise ValueError("Does not recognize object type.")

    if cleaning not in ["v21", "vm21"]:
        raise ValueError("Does not recognize cleaning procedure.")

    if c is None:
        c, unc = position.find_center(ra, dec)

    if r_cut is None:
        ri = angle.sky_distance_deg(ra, dec, c[0], c[1])
        results, var = position.maximum_likelihood(x=np.asarray([ri]), model="plummer")
        r_cut = 10 * 10 ** results[0]

    idx = np.where(angle.sky_distance_deg(ra, dec, c[0], c[1]) < r_cut)

    if err_lim is None:
        ini = pm.initial_guess(pmra[idx], pmdec[idx])
        if object_type == "gc":
            err_lim = 0.5 * ini[2]
        elif object_type == "dsph":
            err_lim = 5 * ini[2]
    else:
        if err_handle == "relative":
            ini = pm.initial_guess(pmra[idx], pmdec[idx])
            err_lim = err_lim * ini[2]

    if cleaning == "v21":

        u = np.sqrt(chi2 / (nu - 5))

        c33 = epmra * epmra
        c34 = epmra * epmdec * corrpm
        c44 = epmdec * epmdec
        err = np.sqrt(
            0.5 * (c33 + c44) + 0.5 * np.sqrt((c44 - c33) ** 2 + 4 * c34 ** 2)
        )

        idx_err = np.where(err < err_lim)
        idx = np.intersect1d(idx, idx_err)

        f = polyRiello20(br_mag)

        C_new = br_excess - f

        sig_c = 0.0059898 + 8.817481 * 1e-12 * g_mag ** (7.618399)

        idx_noise1 = np.where(np.abs(C_new) < 3 * sig_c)
        idx_noise2 = np.where(u < 1.2 * np.maximum(1, np.exp(-0.2 * (g_mag - 19.5))))

        idx_noise = np.intersect1d(idx_noise1, idx_noise2)

        idx = np.intersect1d(idx, idx_noise)

    elif cleaning == "vm21":

        u = np.sqrt(chi2 / (nu - 5))

        c33 = epmra * epmra
        c34 = epmra * epmdec * corrpm
        c44 = epmdec * epmdec
        err = np.sqrt(
            0.5 * (c33 + c44) + 0.5 * np.sqrt((c44 - c33) ** 2 + 4 * c34 ** 2)
        )

        idx_err = np.where(err < err_lim)
        idx = np.intersect1d(idx, idx_err)

        idx_noiseE1 = np.where(1.0 + 0.015 * br_mag ** 2 < br_excess)
        idx_noiseE2 = np.where(br_excess < 1.3 + 0.06 * br_mag ** 2)
        idx_noise1 = np.intersect1d(idx_noiseE1, idx_noiseE2)
        idx_noise2 = np.where(u < 1.2 * np.maximum(1, np.exp(-0.2 * (g_mag - 19.5))))

        idx_noise = np.intersect1d(idx_noise1, idx_noise2)

        idx = np.intersect1d(idx, idx_noise)

    return idx


def find_object(
    ra,
    dec,
    pmra,
    pmdec,
    epmra,
    epmdec,
    corrpm,
    chi2,
    nu,
    g_mag,
    br_mag,
    br_excess,
    sd_model="plummer",
    min_method="dif",
    prob_method="complete",
    prob_limit=0.9,
    use_hrd=True,
    nsig=3,
    bw_hrd=None,
    r_max=None,
    err_lim=None,
    err_handle="relative",
    object_type="gc",
    return_center=False,
    check_fit=False,
    cleaning="v21",
    conv=True,
):

    """
    Finds galactic object by employing the analysis proposed in
    Vitral 2021.

    Parameters
    ----------
    ra : array_like
        Gaia designation: ra.
    dec : array_like
        Gaia designation: dec.
    pmra : array_like
        Gaia designation: pmra.
    pmdec : array_like
        Gaia designation: pmdec.
    epmra : array_like
        Gaia designation: pmra_error.
    epmdec : array_like
        Gaia designation: pmdec_error.
    corrpm : array_like
        Gaia designation: pmra_pmdec_corr.
    chi2 : array_like
        Gaia designation: astrometric_chi2_al.
    nu : array_like
        Gaia designation: astrometric_n_good_obs_al.
    g_mag : array_like
        Gaia designation: phot_g_mean_mag.
    br_mag : array_like
        Gaia designation: bp_rp.
    br_excess : array_like
        Gaia designation: phot_bp_rp_excess_factor.
    sd_model : string, optional
        Surface density model to be considered. Available options are:
             - 'sersic'
             - 'kazantzidis'
             - 'plummer'
             - 'test', for testing which among Plummer and Sersic should be
             used, based on AICc.
        The default is 'plummer'.
    min_method : string, optional
        Minimization method to be used by the pm maximum likelihood fit.
        The default is 'dif'.
    prob_method : string, optional
        Method of probability cut chosen. Available options are:
            - complete: Considers proper motions and projected radii.
            - pm: Considers only proper motions.
            - position: Considers only projected radii.
        The default is 'complete'.
    prob_limit : float, optional
        Probability threshold (between 0 and 1) considered to cut
        the data. The default is 0.9.
    use_hrd : boolean, optional
        True if the user wants to select stars inside a n-sigma contour
        region inside the color-magnitude diagram.
        The default is True.
    nsig : float, optional
        Number of sigma up to when consider stars inside the KDE
        color-magnitude diagram.
        The default is 3.3.
    bw_hrd : float or string, optional
        Bandwidth method of KDE's scipy method. The default is None.
    r_max : float, optional
        Maximum projected radius up to where consider the data.
        The default is None
    err_lim : float, optional
        Error threshold for proper motions. The default is None.
    err_handle : string, optional
        Way in which the error limit is handled:
            - 'relative', if the err_lim argument is the number of times
            the initial guess on the galactic object proper motion dispersion.
            - 'absolute', if the err_lim argument is the absolute error
            limit in mas/yr.
        The default is 'relative'.
    object_type : string, optional
        Galactic object type, to be considered when defining the default
        value of err_lim:
            - 'gc', for globular cluster.
            - 'dsph', for dwarf spheroidal galaxy.
        The default is 'gc'.
    return_center : boolean, optional
        True if the user wants an estimate of the center as output.
        The default is False.
    check_fit : boolean, optional
        True is the user wants to plot validity checks throughout the fitting
        procedure.
        The default is False.
    cleaning : string, optional
        Cleaning procedure to be adopted:
            - 'v21', to use the same cleaning from Vitral 2021
            - 'vm21', to use the same cleaning from Vitral & Mamon 21
        The default is 'v21'.
    conv : boolean, optional
        True, if the user wants to convolve the galactic object PDF with
        Gaussian errors. The defualt is True.

    Raises
    ------
    ValueError
        Surface density model is not one of the following:
            - 'sersic'
            - 'kazantzidis'
            - 'plummer'
            - 'test'
        Probability method is not one of the following:
            - 'complete'
            - 'pm'
            - 'position'
        Probability threshold is not inbetween 0 and 1.
        nsig is not positive.
        err_handle is not one of the following:
            - 'relative'
            - 'absolute'
        object_type is not one of the following:
            - 'gc'
            - 'dsph'
        cleaning is not one of the following:
            - 'v21'
            - 'vm21'

    Returns
    -------
    idx_final : array_like
        Array containing the indexes of the stars in original file who have
        passed the probability cuts.
    results_sd : array_like
        Results of the surface density fit.
    var_sd : array_like
        Uncertainties of the surface density fit
    results_pm : array_like
        Results of the proper motion fit.
    var_pm : array_like
        Uncertainties of the proper motion fit.

    """

    if sd_model not in ["sersic", "plummer", "kazantzidis", "test"]:
        raise ValueError("Does not recognize surface density model.")

    if prob_method not in ["complete", "pm", "position"]:
        raise ValueError("Does not recognize probability model.")

    if prob_limit > 1 or prob_limit < 0:
        raise ValueError("Probability threshold must be inbetween 0 and 1.")

    if nsig <= 0:
        raise ValueError("nsig must be positive.")

    if err_handle not in ["relative", "absolute"]:
        raise ValueError("Does not recognize method to handle errors.")

    if object_type not in ["gc", "dsph"]:
        raise ValueError("Does not recognize object type.")

    if cleaning not in ["v21", "vm21"]:
        raise ValueError("Does not recognize cleaning procedure.")

    c, unc = position.find_center(ra, dec)

    ri = angle.sky_distance_deg(ra, dec, c[0], c[1])
    if r_max is None:
        ri_sd = ri
    else:
        ri_sd = ri[np.where(ri <= r_max)]

    if sd_model == "sersic":
        results_sd, var_sd = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="sersic"
        )
        r_cut = 10 * 10 ** results_sd[1]
        nsys = len(ri_sd) / (1 + 10 ** results_sd[2])
        nilop = len(ri_sd) - nsys
        sd = (
            position.sd_sersic(results_sd[0], sorted(ri_sd) / 10 ** results_sd[1])
            * nsys
            / (np.pi * (10 ** results_sd[1]) ** 2)
        )
        prob_sd = position.prob(ri, results_sd, model="sersic")
    elif sd_model == "plummer":
        results_sd, var_sd = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="plummer"
        )
        r_cut = 10 * 10 ** results_sd[0]
        nsys = len(ri_sd) / (1 + 10 ** results_sd[1])
        nilop = len(ri_sd) - nsys
        sd = (
            position.sd_plummer(sorted(ri_sd) / 10 ** results_sd[0])
            * nsys
            / (np.pi * (10 ** results_sd[0]) ** 2)
        )
        prob_sd = position.prob(ri, results_sd, model="plummer")
    elif sd_model == "kazantzidis":
        results_sd, var_sd = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="kazantzidis"
        )
        r_cut = 10 * 10 ** results_sd[0]
        nsys = len(ri_sd) / (1 + 10 ** results_sd[1])
        nilop = len(ri_sd) - nsys
        sd = (
            position.sd_kazantzidis(sorted(ri_sd) / 10 ** results_sd[0])
            * nsys
            / (np.pi * (10 ** results_sd[0]) ** 2)
        )
        prob_sd = position.prob(ri, results_sd, model="kazantzidis")
    elif sd_model == "test":
        results_sd_s, var_sd_s = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="sersic"
        )
        results_sd_p, var_sd_p = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="plummer"
        )
        results_sd_k, var_sd_k = position.maximum_likelihood(
            x=np.asarray([ri_sd]), model="kazantzidis"
        )

        aicc_s = get_aicc(
            position.likelihood_sersic(results_sd_s, ri_sd), 3, len(ri_sd)
        )
        aicc_p = get_aicc(
            position.likelihood_plummer(results_sd_p, ri_sd), 2, len(ri_sd)
        )
        aicc_k = get_aicc(
            position.likelihood_kazantzidis(results_sd_k, ri_sd), 2, len(ri_sd)
        )

        delta_aicc = aicc_p - aicc_s

        if delta_aicc < 2:
            delta_aicc = aicc_p - aicc_k
            if delta_aicc < 2:
                sd_model = "plummer"
                results_sd, var_sd = results_sd_p, var_sd_p
                r_cut = 10 * 10 ** results_sd[0]
                nsys = len(ri_sd) / (1 + 10 ** results_sd[1])
                nilop = len(ri_sd) - nsys
                sd = (
                    position.sd_plummer(sorted(ri_sd) / 10 ** results_sd[0])
                    * nsys
                    / (np.pi * (10 ** results_sd[0]) ** 2)
                )
                prob_sd = position.prob(ri, results_sd, model="plummer")
            else:
                sd_model = "kazantzidis"
                results_sd, var_sd = results_sd_k, var_sd_k
                r_cut = 10 * 10 ** results_sd[0]
                nsys = len(ri_sd) / (1 + 10 ** results_sd[1])
                nilop = len(ri_sd) - nsys
                sd = (
                    position.sd_kazantzidis(sorted(ri_sd) / 10 ** results_sd[0])
                    * nsys
                    / (np.pi * (10 ** results_sd[0]) ** 2)
                )
                prob_sd = position.prob(ri, results_sd, model="kazantzidis")
        else:
            delta_aicc = aicc_k - aicc_s
            if delta_aicc < 2:
                sd_model = "kazantzidis"
                results_sd, var_sd = results_sd_k, var_sd_k
                r_cut = 10 * 10 ** results_sd[0]
                nsys = len(ri_sd) / (1 + 10 ** results_sd[1])
                nilop = len(ri_sd) - nsys
                sd = (
                    position.sd_kazantzidis(sorted(ri_sd) / 10 ** results_sd[0])
                    * nsys
                    / (np.pi * (10 ** results_sd[0]) ** 2)
                )
                prob_sd = position.prob(ri, results_sd, model="kazantzidis")
            else:
                sd_model = "sersic"
                results_sd, var_sd = results_sd_s, var_sd_s
                r_cut = 10 * 10 ** results_sd[1]
                nsys = len(ri_sd) / (1 + 10 ** results_sd[2])
                nilop = len(ri_sd) - nsys
                sd = (
                    position.sd_sersic(
                        results_sd[0], sorted(ri_sd) / 10 ** results_sd[1]
                    )
                    * nsys
                    / (np.pi * (10 ** results_sd[1]) ** 2)
                )
                prob_sd = position.prob(ri, results_sd, model="sersic")

    if check_fit is True:
        surf_dens = position.surface_density(x=ri_sd)

        fs = nilop / (np.pi * np.amax(ri_sd) ** 2)

        plt.title(sd_model)
        plt.bar(
            surf_dens[0],
            surf_dens[1],
            surf_dens[3],
            ec="k",
            color=(102 / 255, 178 / 255, 1),
            alpha=0.3,
        )
        plt.xscale("log")
        plt.yscale("log")
        plt.ylabel("Surface density")
        plt.xlabel("Projected radii")
        plt.errorbar(
            surf_dens[0],
            surf_dens[1],
            color="black",
            yerr=surf_dens[2],
            capsize=3,
            ls="none",
            barsabove=True,
            zorder=11,
        )
        plt.plot(sorted(ri_sd), sd + fs, lw=3, color="red")
        plt.show()

    if r_max is None:
        r_max = r_cut

    idx = clean_gaia(
        pmra,
        pmdec,
        epmra,
        epmdec,
        corrpm,
        chi2,
        nu,
        g_mag,
        br_mag,
        br_excess,
        ra,
        dec,
        c=c,
        r_cut=r_max,
        err_lim=err_lim,
        err_handle=err_handle,
        object_type=object_type,
        cleaning=cleaning,
    )

    prob_sd = prob_sd[idx]
    pmra = pmra[idx]
    pmdec = pmdec[idx]
    epmra = epmra[idx]
    epmdec = epmdec[idx]
    corrpm = corrpm[idx]

    results_pm, var_pm = pm.maximum_likelihood(
        pmra, pmdec, eX=epmra, eY=epmdec, eXY=corrpm, min_method=min_method, conv=conv
    )

    prob_pm = pm.prob(pmra, pmdec, epmra, epmdec, corrpm, results_pm, conv=conv)

    if check_fit is True:
        ellipse_rot = angle.get_ellipse(
            results_pm[5], results_pm[6], results_pm[7], 200
        )
        circle = angle.get_ellipse(results_pm[2], results_pm[2], 0, 200)

        ranges = [
            [pm.quantile(pmra, 0.025)[0], pm.quantile(pmra, 0.975)[0]],
            [pm.quantile(pmdec, 0.025)[0], pm.quantile(pmdec, 0.975)[0]],
        ]

        plt.xlabel("pmra")
        plt.ylabel("pmdec")
        plt.xlim(ranges[0][0], ranges[0][1])
        plt.ylim(ranges[1][0], ranges[1][1])
        plt.hist2d(
            pmra,
            pmdec,
            range=ranges,
            bins=(position.good_bin(pmra), position.good_bin(pmdec)),
            cmap="hot",
            norm=mc.LogNorm(),
        )
        plt.plot(
            results_pm[3] + ellipse_rot[0, :],
            results_pm[4] + ellipse_rot[1, :],
            "green",
            lw=3,
        )
        plt.plot(
            results_pm[0] + circle[0, :], results_pm[1] + circle[1, :], "blue", lw=3
        )
        plt.show()

        x, y = angle.rotate_axis(
            pmra, pmdec, results_pm[7], results_pm[3], results_pm[4]
        )
        mu_x, mu_y = angle.rotate_axis(
            results_pm[0], results_pm[1], results_pm[7], results_pm[3], results_pm[4]
        )

        ranges = [
            [pm.quantile(x, 0.025)[0], pm.quantile(x, 0.975)[0]],
            [pm.quantile(y, 0.025)[0], pm.quantile(y, 0.975)[0]],
        ]

        Ux = np.linspace(ranges[0][0], ranges[0][1], 200)
        Uy = np.linspace(ranges[1][0], ranges[1][1], 200)

        projx = pm.proj_global_pdf(
            Ux, mu_x, results_pm[2], results_pm[5], results_pm[8], results_pm[9]
        )
        projy = pm.proj_global_pdf(
            Uy, mu_y, results_pm[2], results_pm[6], results_pm[8], results_pm[9]
        )

        plt.title("PMs projected on semi-major axis")
        plt.hist(
            x,
            100,
            range=ranges[0],
            alpha=1,
            density=True,
            histtype="stepfilled",
            color=(102 / 255, 178 / 255, 1),
        )
        plt.plot(Ux, projx, color="darkgreen", lw=3)
        plt.xlim(ranges[0][0], ranges[0][1])
        plt.show()

        plt.title("PMs projected on semi-minor axis")
        plt.hist(
            y,
            100,
            range=ranges[1],
            alpha=1,
            density=True,
            histtype="stepfilled",
            color=(102 / 255, 178 / 255, 1),
        )
        plt.plot(Uy, projy, color="darkgreen", lw=3)
        plt.xlim(ranges[1][0], ranges[1][1])
        plt.show()

    if prob_method == "complete":
        probs = prob_pm * prob_sd
    elif prob_method == "pm":
        probs = prob_pm
    elif prob_method == "position":
        probs = prob_sd

    idx_p = np.where(probs > prob_limit)

    if use_hrd is True:
        probs = probs[idx_p]
        fg_mag = g_mag[idx][idx_p]
        fbr_mag = br_mag[idx][idx_p]

        if bw_hrd is None:
            bw_hrd = hrd.bw_silver(fbr_mag, fg_mag) * 0.5

        z = hrd.kde(fbr_mag, fg_mag, ww=probs, bw=bw_hrd)

        max_z = np.amax(z)

        levels = max_z * np.exp(-0.5 * np.array([nsig]) ** 2)

        qx_i, qx_f = np.nanmin(fbr_mag), np.nanmax(fbr_mag)
        qy_i, qy_f = np.nanmin(fg_mag), np.nanmax(fg_mag)
        ranges = [[qx_i, qx_f], [qy_i, qy_f]]

        if check_fit is True:

            fig, ax = plt.subplots(figsize=(7, 6))
            plt.title(r"Isochrone", fontsize=16)
            plt.xlabel(r"color", fontsize=15)
            plt.ylabel(r"main magnitude", fontsize=15)

            cax = ax.imshow(
                z,
                origin="lower",
                aspect="auto",
                extent=[ranges[0][0], ranges[0][1], ranges[1][0], ranges[1][1]],
                cmap="hot_r",
                norm=mc.LogNorm(vmin=0.001),
            )

            cbar = fig.colorbar(cax)
            cbar.ax.set_title("KDE's PDF", fontsize=15)

            contours = plt.contour(
                z,
                origin="lower",
                levels=levels,
                extent=[ranges[0][0], ranges[0][1], ranges[1][0], ranges[1][1]],
            )
            plt.gca().invert_yaxis()
            plt.show()

        else:

            contours = plt.contour(
                z,
                origin="lower",
                levels=levels,
                extent=[ranges[0][0], ranges[0][1], ranges[1][0], ranges[1][1]],
            )
            plt.clf()

        idx_final = hrd.inside_contour(br_mag, g_mag, contours, idx[idx_p])

    else:

        idx_final = idx_p

    if return_center is False:
        return idx_final, results_sd, var_sd, results_pm, var_pm
    else:
        return idx_final, results_sd, var_sd, results_pm, var_pm, c


def extract_object(
    path,
    sd_model="plummer",
    min_method="dif",
    prob_method="complete",
    prob_limit=0.9,
    use_hrd=True,
    nsig=3,
    bw_hrd=None,
    r_max=None,
    err_lim=None,
    err_handle="relative",
    object_type="gc",
    return_center=False,
    check_fit=False,
    cleaning="v21",
    conv=True,
):
    """


    Parameters
    ----------
    path : string
        Path to Gaia fits file.
    sd_model : string, optional
        Surface density model to be considered. Available options are:
             - 'sersic'
             - 'kazantzidis'
             - 'plummer'
             - 'test', for testing which among Plummer and Sersic should be
             used, based on AICc.
        The default is 'plummer'.
    min_method : string, optional
        Minimization method to be used by the pm maximum likelihood fit.
        The default is 'dif'.
    prob_method : string, optional
        Method of probability cut chosen. Available options are:
            - complete: Considers proper motions and projected radii.
            - pm: Considers only proper motions.
            - position: Considers only projected radii.
        The default is 'complete'.
    prob_limit : float, optional
        Probability threshold (between 0 and 1) considered to cut
        the data. The default is 0.9.
    use_hrd : boolean, optional
        True if the user wants to select stars inside a n-sigma contour
        region inside the color-magnitude diagram.
        The default is True.
    nsig : float, optional
        Number of sigma up to when consider stars inside the KDE
        color-magnitude diagram.
        The default is 3.3.
    bw_hrd : float or string, optional
        Bandwidth method of KDE's scipy method. The default is None.
    r_max : float, optional
        Maximum projected radius up to where consider the data.
        The default is None
    err_lim : float, optional
        Error threshold for proper motions. The default is None.
    err_handle : string, optional
        Way in which the error limit is handled:
            - 'relative', if the err_lim argument is the number of times
            the initial guess on the galactic object proper motion dispersion.
            - 'absolute', if the err_lim argument is the absolute error
            limit in mas/yr.
        The default is 'relative'.
    object_type : string, optional
        Galactic object type, to be considered when defining the default
        value of err_lim:
            - 'gc', for globular cluster.
            - 'dsph', for dwarf spheroidal galaxy.
        The default is 'gc'.
    return_center : boolean, optional
        True if the user wants an estimate of the center as output.
        The default is False.
    check_fit : boolean, optional
        True is the user wants to plot validity checks throughout the fitting
        procedure.
        The default is False.
    cleaning : string, optional
        Cleaning procedure to be adopted:
            - 'v21', to use the same cleaning from Vitral 2021
            - 'vm21', to use the same cleaning from Vitral & Mamon 21
        The default is 'v21'.
    conv : boolean, optional
        True, if the user wants to convolve the galactic object PDF with
        Gaussian errors. The defualt is True.


    Raises
    ------
    ValueError
        Surface density model is not one of the following:
            - 'sersic'
            - 'kazantzidis'
            - 'plummer'
            - 'test'
        Probability method is not one of the following:
            - 'complete'
            - 'pm'
            - 'position'
        Probability threshold is not inbetween 0 and 1.
        nsig is not positive.
        err_handle is not one of the following:
            - 'relative'
            - 'absolute'
        object_type is not one of the following:
            - 'gc'
            - 'dsph'
        cleaning is not one of the following:
            - 'v21'
            - 'vm21'

    Returns
    -------
    full_data : array_like
        File containing the information of the stars that passed the
        probability cuts.
    results_sd : array_like
        Results of the surface density fit.
    var_sd : array_like
        Uncertainties of the surface density fit
    results_pm : array_like
        Results of the proper motion fit.
    var_pm : array_like
        Uncertainties of the proper motion fit.

    """

    if sd_model not in ["sersic", "plummer", "kazantzidis", "test"]:
        raise ValueError("Does not recognize surface density model.")

    if prob_method not in ["complete", "pm", "position"]:
        raise ValueError("Does not recognize probability model.")

    if prob_limit > 1 or prob_limit < 0:
        raise ValueError("Probability threshold must be inbetween 0 and 1.")

    if nsig <= 0:
        raise ValueError("nsig must be positive.")

    if err_handle not in ["relative", "absolute"]:
        raise ValueError("Does not recognize method to handle errors.")

    if object_type not in ["gc", "dsph"]:
        raise ValueError("Does not recognize object type.")

    if cleaning not in ["v21", "vm21"]:
        raise ValueError("Does not recognize cleaning procedure.")

    hdu = fits.open(path)

    ra = np.asarray(hdu[1].data[:]["ra      "])  # degrees
    dec = np.asarray(hdu[1].data[:]["DEC     "])  # degrees
    pmra = np.asarray(hdu[1].data[:]["pmra    "])  # mas/yr
    epmra = np.asarray(hdu[1].data[:]["pmra_error"])  # mas/yr
    pmdec = np.asarray(hdu[1].data[:]["pmdec   "])  # mas/yr
    epmdec = np.asarray(hdu[1].data[:]["pmdec_error"])  # mas/yr
    corrpm = np.asarray(hdu[1].data[:]["pmra_pmdec_corr"])  # mas/yr
    g_mag = np.asarray(hdu[1].data[:]["phot_g_mean_mag"])  # in mag
    br_mag = np.asarray(hdu[1].data[:]["bp_rp   "])  # in mag
    chi2 = np.asarray(hdu[1].data[:]["astrometric_chi2_al"])  # no units
    nu = np.asarray(hdu[1].data[:]["astrometric_n_good_obs_al"])  # no units
    br_excess = np.asarray(hdu[1].data[:]["phot_bp_rp_excess_factor"])  # no units

    full_data = hdu[1].data[:]

    hdu.close()

    if return_center is False:
        idx_final, results_sd, var_sd, results_pm, var_pm = find_object(
            ra,
            dec,
            pmra,
            pmdec,
            epmra,
            epmdec,
            corrpm,
            chi2,
            nu,
            g_mag,
            br_mag,
            br_excess,
            sd_model=sd_model,
            min_method=min_method,
            prob_method=prob_method,
            prob_limit=prob_limit,
            use_hrd=use_hrd,
            nsig=nsig,
            bw_hrd=bw_hrd,
            r_max=r_max,
            err_lim=err_lim,
            err_handle=err_handle,
            object_type=object_type,
            check_fit=check_fit,
            cleaning=cleaning,
            conv=conv,
        )
    else:
        idx_final, results_sd, var_sd, results_pm, var_pm, center = find_object(
            ra,
            dec,
            pmra,
            pmdec,
            epmra,
            epmdec,
            corrpm,
            chi2,
            nu,
            g_mag,
            br_mag,
            br_excess,
            sd_model=sd_model,
            min_method=min_method,
            prob_method=prob_method,
            prob_limit=prob_limit,
            use_hrd=use_hrd,
            nsig=nsig,
            bw_hrd=bw_hrd,
            r_max=r_max,
            err_lim=err_lim,
            err_handle=err_handle,
            object_type=object_type,
            return_center=return_center,
            check_fit=check_fit,
            cleaning=cleaning,
            conv=conv,
        )

    full_data = full_data[idx_final]

    if return_center is False:
        return full_data, results_sd, var_sd, results_pm, var_pm
    else:
        return full_data, results_sd, var_sd, results_pm, var_pm, center


# %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# ---------------------------------------------------------------------------
"Statistical handling"
# ---------------------------------------------------------------------------


def get_aic(lnL, Nf):
    """
    Provides the Akaike Information Criterion (Akaike, 1973), i.e. AIC.

    Parameters
    ----------
    lnL : float
        Minus the log-likelihood.
    Nf : int
        Number of free parameters.

    Returns
    -------
    AIC : float
        AIC indicator.

    """

    AIC = 2 * (lnL + Nf)
    return AIC


def get_bic(lnL, Nf, Nd):
    """
    Provides the Bayes Information Criterion (Schwarz, 1978), i.e. BIC.

    Parameters
    ----------
    lnL : float
        Minus the log-likelihood.
    Nf : int
        Number of free parameters.
    Nd : int
        Size of data sample.

    Returns
    -------
    BIC : float
        BIC indicator.

    """

    BIC = 2 * lnL + Nf * np.log(Nd)
    return BIC


def get_aicc(lnL, Nf, Nd):
    """
    Provides the corrected Akaike Information Criterion (Sugiyara, 1978),
    i.e. AICc.

    Parameters
    ----------
    lnL : float
        Minus the log-likelihood.
    Nf : int
        Number of free parameters.
    Nd : int
        Size of data sample.

    Returns
    -------
    AICc : float
        AICc indicator.

    """

    AICc = get_aic(lnL, Nf) + 2 * Nf * (1 + Nf) / (Nd - Nf - 1)
    return AICc
